package sun.nio.ch;

import java.nio.ByteBuffer;
import java.util.Iterator;
import java.util.concurrent.Future;
import java.nio.channels.FileLock;
import java.nio.channels.CompletionHandler;
import java.nio.channels.AsynchronousCloseException;
import java.io.IOException;
import java.nio.channels.ClosedChannelException;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.concurrent.ExecutorService;
import java.io.FileDescriptor;
import java.util.concurrent.locks.ReadWriteLock;
import java.nio.channels.AsynchronousFileChannel;

abstract class AsynchronousFileChannelImpl extends AsynchronousFileChannel
{
    protected final ReadWriteLock closeLock;
    protected volatile boolean closed;
    protected final FileDescriptor fdObj;
    protected final boolean reading;
    protected final boolean writing;
    protected final ExecutorService executor;
    private volatile FileLockTable fileLockTable;
    
    protected AsynchronousFileChannelImpl(FileDescriptor fdObj, boolean reading, boolean writing, ExecutorService executor) {
        this.closeLock = new ReentrantReadWriteLock();
        this.fdObj = fdObj;
        this.reading = reading;
        this.writing = writing;
        this.executor = executor;
    }
    
    final ExecutorService executor() {
        return this.executor;
    }
    
    @Override
    public final boolean isOpen() {
        return !this.closed;
    }
    
    protected final void begin() throws IOException {
        this.closeLock.readLock().lock();
        if (this.closed) {
            throw new ClosedChannelException();
        }
        return;
    }
    
    protected final void end() {
        this.closeLock.readLock().unlock();
    }
    
    protected final void end(boolean b) throws IOException {
        this.end();
        if (!b && !this.isOpen()) {
            throw new AsynchronousCloseException();
        }
        return;
    }
    
    abstract <A> Future<FileLock> implLock(long p0, long p1, boolean p2, A p3, CompletionHandler<FileLock, ? super A> p4);
    
    @Override
    public final Future<FileLock> lock(long n, long n2, boolean b) {
        return this.implLock(n, n2, b, null, null);
    }
    
    @Override
    public final <A> void lock(long n, long n2, boolean b, A a, CompletionHandler<FileLock, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        this.implLock(n, n2, b, a, completionHandler);
        return;
    }
    
    final void ensureFileLockTableInitialized() throws IOException {
        if (this.fileLockTable == null) {
            synchronized (this) {
                if (this.fileLockTable == null) {
                    this.fileLockTable = FileLockTable.newSharedFileLockTable(this, this.fdObj);
                }
            }
        }
    }
    
    final void invalidateAllLocks() throws IOException {
        if (this.fileLockTable != null) {
            for (FileLock fileLock : this.fileLockTable.removeAll()) {
                synchronized (fileLock) {
                    if (fileLock.isValid()) {
                        FileLockImpl fileLockImpl = (FileLockImpl)fileLock;
                        this.implRelease(fileLockImpl);
                        fileLockImpl.invalidate();
                    }
                    else {
                        continue;
                    }
                }
            }
        }
    }
    
    protected final FileLockImpl addToFileLockTable(long n, long n2, boolean b) {
        FileLockImpl fileLockImpl;
        try {
            this.closeLock.readLock().lock();
            if (this.closed) {
                return null;
            }
            try {
                this.ensureFileLockTableInitialized();
            }
            catch (IOException detailMessage) {
                throw new AssertionError(detailMessage);
            }
            fileLockImpl = new FileLockImpl(this, n, n2, b);
            this.fileLockTable.add(fileLockImpl);
        }
        finally {
            this.end();
        }
        return fileLockImpl;
    }
    
    protected final void removeFromFileLockTable(FileLockImpl fileLockImpl) {
        this.fileLockTable.remove(fileLockImpl);
    }
    
    protected abstract void implRelease(FileLockImpl p0) throws IOException;
    
    final void release(FileLockImpl fileLockImpl) throws IOException {
        try {
            this.begin();
            this.implRelease(fileLockImpl);
            this.removeFromFileLockTable(fileLockImpl);
        }
        finally {
            this.end();
        }
    }
    
    abstract <A> Future<Integer> implRead(ByteBuffer p0, long p1, A p2, CompletionHandler<Integer, ? super A> p3);
    
    @Override
    public final Future<Integer> read(ByteBuffer byteBuffer, long n) {
        return this.implRead(byteBuffer, n, null, null);
    }
    
    @Override
    public final <A> void read(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        this.implRead(byteBuffer, n, a, completionHandler);
        return;
    }
    
    abstract <A> Future<Integer> implWrite(ByteBuffer p0, long p1, A p2, CompletionHandler<Integer, ? super A> p3);
    
    @Override
    public final Future<Integer> write(ByteBuffer byteBuffer, long n) {
        return this.implWrite(byteBuffer, n, null, null);
    }
    
    @Override
    public final <A> void write(ByteBuffer byteBuffer, long n, A a, CompletionHandler<Integer, ? super A> completionHandler) {
        if (completionHandler == null) {
            throw new NullPointerException("'handler' is null");
        }
        this.implWrite(byteBuffer, n, a, completionHandler);
        return;
    }
}
